import mimetypes
import logging
import os
from typing import Dict, Any

from .activity import BaseActivity, ActivityError
from ..types import ActivityJob, JobState
from .readers import PdfReader, FileReader


class ReadActivity(BaseActivity):
    def __init__(self, pdf_reader: PdfReader, file_reader: FileReader):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._handlers = {
            "application/pdf": pdf_reader,
            "application/json": file_reader,
            "text/plain": file_reader,
            "text/markdown": file_reader
        }

    async def perform(self, activity_job: ActivityJob, inputs: Dict[str, Any]) -> None:
        try:
            file_path = inputs.get("file_path")
            if not file_path or not os.path.exists(file_path):
                raise ActivityError(f"Invalid file path: {file_path}")

            content_type = mimetypes.guess_type(file_path)[0]
            handler = self._handlers.get(content_type)
            if handler:
                content = await handler.read(file_path)
            else:
                raise ActivityError(f"Cannot read content of type: {content_type}")

            # Add the file path to the app context
            app_job = activity_job.app_job
            app_job.context["file_path"] = file_path

            activity_job.state = JobState.SUCCESS
            activity_job.output = content
            activity_job.output_type = "text/plain"

        except Exception as e:
            activity_job.state = JobState.ERROR
            activity_job.output = str(e)
            self._logger.error(f"Encountered an error {e}")



import json
import logging
from typing import Dict, Any, List

from .activity import BaseActivity, ActivityError
from ..providers import ProviderManager
from ..providers.types import (
    ChatCompletionMessage,
    ChatCompletionRequest,
    Choice,
    Function,
    Tool
)
from ..types import ActivityJob, JobState


class ExtractActivity(BaseActivity):
    def __init__(self, provider_manager: ProviderManager):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._provider_manager = provider_manager

    async def perform(self, activity_job: ActivityJob, inputs: Dict[str, Any]) -> None:
        try:
            app = activity_job.app_job.app
            activity = app.activities[activity_job.activity_name]

            if len(activity.models) < 1:
                raise ActivityError("A model is required")
            model = app.models[activity.models[0]]

            schema = activity.parameters.get("schema", None)
            if not schema:
                raise ActivityError("A schema is required")
            if not isinstance(schema, list):
                raise ActivityError("A schema must be a list of types")

            tools = []
            for schema_def in schema:
                func_def = Function(**schema_def)
                tools.append(Tool(function=func_def))

            messages = []
            profile = app.info.profile
            if profile:
                messages.append(ChatCompletionMessage(role="system", content=profile))

            messages.append(ChatCompletionMessage(role="system", content="""
                Use the tools provided with information extracted from the user prompt.
                Use the entire prompt as the source of information. Do not exclude or omit anything.
            """))
            messages.append(ChatCompletionMessage(role="user", content=self.merge_inputs(inputs)))

            values: Dict[str, List[Any]] = {}
            provider = self._provider_manager.get_provider(model.provider)
            for x in range(self.MAX_ITERATIONS):
                request = ChatCompletionRequest(
                    model=model.model,
                    messages=messages,
                    temperature=0.0,
                    tools=tools,
                    tool_choice="auto"
                )
                response = await provider.create_completion(request)

                choice: Choice = response.choices[0]
                message: ChatCompletionMessage = choice.message
                messages.append(message)

                if choice.finish_reason == "tool_calls":
                    for tool_call in message.tool_calls:
                        function_call = tool_call.function
                        values_for_type = values.get(function_call.name, None)
                        if not values_for_type:
                            values_for_type = []
                            values[function_call.name] = values_for_type
                        values_for_type.append(json.loads(function_call.arguments))
                        messages.append(ChatCompletionMessage(
                            role="tool",
                            tool_call_id=tool_call.id,
                            name=function_call.name,
                            content="Success"
                        ))
                elif choice.finish_reason:
                    break

            activity_job.state = JobState.SUCCESS
            activity_job.output = json.dumps(values)
            activity_job.output_type = "application/json"

        except Exception as e:
            self._logger.error(e)
            activity_job.state = JobState.ERROR
            activity_job.output = str(e)





import logging
from typing import Dict, Any, List

from openapi_parser import parse
from openapi_parser.enumeration import OperationMethod

from ..tool import BaseTool, ToolError
from ...http_client import AsyncHttpClient
from ...providers.types import Tool
from ...types import ToolDef


class RestTool(BaseTool):
    def __init__(self, http_client: AsyncHttpClient):
        self._http_client = http_client
        logging.getLogger('openapi_parser').setLevel(logging.ERROR)
        self._logger = logging.getLogger(self.__class__.__name__)

    async def get_metadata(self, tool_def: ToolDef) -> List[Tool]:
        params = tool_def.parameters or {}
        endpoint = params.get("endpoint", None)
        if not endpoint:
            raise ToolError(f"An REST service endpoint is required")

        openapi_url = params.get("openapi", self._http_client.urljoin(endpoint, "openapi.json"))
        try:
            tools = []
            data = await self._http_client.get(openapi_url)
            spec = parse(spec_string=data)
            for path in spec.paths:
                for operation in path.operations:
                    if operation.method == OperationMethod.GET:
                        required = []
                        props = {}
                        for parameter in operation.parameters:
                            if parameter.required:
                                required.append(parameter.name)
                            props[parameter.name] = {
                                "type": parameter.schema.type.value,
                                "description": parameter.description
                            }
                        path_steps = path.url.split("/")
                        tool = {
                            "function": {
                                "name": path_steps[1],
                                "description": operation.description,
                                "parameters": {
                                    "properties": props,
                                    "required": required
                                }
                            }
                        }
                        tools.append(Tool(**tool))
            return tools
        except Exception as e:
            self._logger.error(f"Failed to get REST metadata {e}")

    async def invoke(self, function_name: str, arguments: Dict[str, Any], tool_def: ToolDef) -> str:
        params = tool_def.parameters or {}
        endpoint = params.get("endpoint", None)
        if not endpoint:
            raise ToolError(f"An REST service endpoint is required")

        url = self._http_client.urljoin(endpoint, function_name)
        return await self._http_client.get(url, arguments, headers={}, json=False)





from .reader import Reader, ReaderError
import aiofiles


class FileReader(Reader):
    async def read(self, file_path: str) -> str:
        try:
            async with aiofiles.open(file_path, mode='r') as f:
                contents = await f.read()
            return contents
        except Exception as e:
            raise ReaderError(f"Failed to read text from a file: {e}")



import os
import logging
import sys

from dependency_injector import containers, providers
from dotenv import load_dotenv, find_dotenv

from .broker import SemanticBroker
from .jobs import JobManager, JobScheduler
from .activities import (
    ReadActivity,
    WriteActivity,
    SummarizeActivity,
    GenerateActivity,
    ExtractActivity,
    StoreActivity,
    RetrieveActivity,
    FunctionActivity,
    ReturnActivity
)
from .activities.readers import PdfReader, FileReader
from .providers import OpenAIProvider, AzureProvider, AnthropicProvider, LlavaProvider
from .http_client import AsyncHttpClient
from .providers import ProviderManager
from .tools.web import WebTool
from .tools.rest import RestTool
from .tools import ToolManager
from .memory import MemoryManager
from .memory.chromadb import ChromaDbRepository


class Container(containers.DeclarativeContainer):
    config = providers.Configuration()

    logging = providers.Resource(
        logging.basicConfig,
        stream=sys.stdout,
        level=config.log.level,
        format=config.log.format,
    )

    http_client = providers.Singleton(
        AsyncHttpClient
    )

    openai_provider = providers.Singleton(
        OpenAIProvider,
        config=config.providers.openai,
        http_client=http_client
    )

    azure_provider = providers.Singleton(
        AzureProvider,
        config=config.providers.azure,
        http_client=http_client
    )

    anthropic_provider = providers.Singleton(
        AnthropicProvider,
        config=config.providers.anthropic,
        http_client=http_client
    )

    llava_provider = providers.Singleton(
        LlavaProvider,
        config=config.providers.llava,
        http_client=http_client
    )

    provider_manager = providers.Singleton(
        ProviderManager,
        openai_provider=openai_provider,
        azure_provider=azure_provider,
        anthropic_provider=anthropic_provider,
        llava_provider=llava_provider
    )

    web_tool = providers.Singleton(
        WebTool,
        config=config.tools.web,
        http_client=http_client
    )

    rest_tool = providers.Singleton(
        RestTool,
        http_client=http_client
    )

    tool_manager = providers.Singleton(
        ToolManager,
        web_tool=web_tool,
        rest_tool=rest_tool
    )

    pdf_reader = providers.Singleton(PdfReader)
    file_reader = providers.Singleton(FileReader)

    read_activity = providers.Singleton(
        ReadActivity,
        file_reader=file_reader,
        pdf_reader=pdf_reader)

    chromadb_repository = providers.Singleton(
        ChromaDbRepository,
        config=config.memory.chromadb
    )

    memory_manager = providers.Singleton(
        MemoryManager,
        chromadb_repository=chromadb_repository
    )

    store_activity = providers.Singleton(
        StoreActivity,
        memory_manager=memory_manager
    )

    retrieve_activity = providers.Singleton(
        RetrieveActivity,
        memory_manager=memory_manager
    )

    write_activity = providers.Singleton(WriteActivity)

    summarize_activity = providers.Singleton(
        SummarizeActivity,
        provider_manager=provider_manager
    )

    generate_activity = providers.Singleton(
        GenerateActivity,
        provider_manager=provider_manager,
        tool_manager=tool_manager
    )

    extract_activity = providers.Singleton(
        ExtractActivity,
        provider_manager=provider_manager
    )

    function_activity = providers.Singleton(FunctionActivity)

    return_activity = providers.Singleton(ReturnActivity)

    job_manager = providers.Singleton(JobManager)

    job_scheduler = providers.Singleton(
        JobScheduler,
        config=config.scheduler,
        job_manager=job_manager,
        read_activity=read_activity,
        write_activity=write_activity,
        summarize_activity=summarize_activity,
        generate_activity=generate_activity,
        extract_activity=extract_activity,
        store_activity=store_activity,
        retrieve_activity=retrieve_activity,
        function_activity=function_activity,
        return_activity=return_activity
    )

    broker = providers.Singleton(
        SemanticBroker,
        job_manager=job_manager,
        job_scheduler=job_scheduler
    )


def get_broker() -> SemanticBroker:
    root = os.path.dirname(find_dotenv())
    load_dotenv()

    container = Container()
    container.config.from_yaml(os.path.join(root, "config.yml"))
    container.init_resources()
    container.wire(modules=[__name__])

    return container.broker()





import logging
from typing import Dict, Any

from pydantic import BaseModel

from ..provider import BaseProvider
from ..types import ChatCompletionRequest, ChatCompletionResponse, Choice, ChatCompletionMessage
from ...http_client import AsyncHttpClient


class LlavaCompletionRequest(BaseModel):
    prompt: str
    temperature: float
    n_predict: int
    cache_prompt: bool = False


class LlavaCompletionResponse(BaseModel):
    content: str
    stop: bool
    stopped_limit: bool


# A provider for llava-v1.5 available here
# https://huggingface.co/jartine/llava-v1.5-7B-GGUF/blob/main/llava-v1.5-7b-q4-server.llamafile
class LlavaProvider(BaseProvider):
    def __init__(self, config: Dict[str, Any], http_client: AsyncHttpClient):
        self._config = config
        self._http_client = http_client
        self._logger = logging.getLogger(self.__class__.__name__)

    async def create_completion(self, request: ChatCompletionRequest) -> ChatCompletionResponse:
        # Format the prompt
        prompts = []
        for message in request.messages:
            if message.role == "user" or message.role == "system":
                prompts.append(f"User: {message.content}")
            elif message.role == "assistant":
                prompts.append(f"Llama: {message.content}")
            else:
                self._logger.error(f"Unknown role {message.role}")

        prompts.append("Llama:")
        prompt = "\n\n" + "\n".join(prompts)

        # Map and send the request
        a_request = LlavaCompletionRequest(prompt=prompt,
                                           temperature=request.temperature,
                                           n_predict=request.max_tokens)
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        url = self._http_client.urljoin(self._config["endpoint"], "/completion")
        data = await self._http_client.post(url, headers, a_request.model_dump_json(exclude_none=True))

        # Map and return the response
        a_response = LlavaCompletionResponse(**data)
        finish_reason = "stop"
        if a_response.stopped_limit:
            finish_reason = "length"
        message = ChatCompletionMessage(role="assistant", content=a_response.content)
        choice = Choice(index=0, message=message, finish_reason=finish_reason)
        response = ChatCompletionResponse(id="", object="object", created=0, choices=[choice])

        return response



from .provider import BaseProvider, ProviderError
from .llava import LlavaProvider
from .azure import AzureProvider
from .openai import OpenAIProvider
from .anthropic import AnthropicProvider
from .manager import ProviderManager

__all__ = [
    "BaseProvider",
    "ProviderManager",
    "OpenAIProvider",
    "AzureProvider",
    "AnthropicProvider",
    "LlavaProvider",
    "ProviderError"
]



import logging

from typing import List, Dict

import yaml

from .jobs import JobManager, JobScheduler
from .types import App


class SemanticBroker:
    def __init__(self, job_manager: JobManager, job_scheduler: JobScheduler):
        self._job_manager = job_manager
        self._job_scheduler = job_scheduler
        self._logger = logging.getLogger(self.__class__.__name__)

    async def run(self, app_file: str, activity_name: str, file_path: str) -> Dict[str, List[str]]:
        # Load the app definition
        with open(app_file) as def_file:
            app_def = yaml.safe_load(def_file)
        app = App(**app_def)
        self._logger.info(f"Loaded {app_file}")

        # Create an activity job
        app_job = self._job_manager.create_app_job(app)
        activity_job = self._job_manager.create_activity_job(app_job, activity_name)

        # Launch this job with the file as an input
        await self._job_scheduler.start_workers()
        await self._job_scheduler.schedule(activity_job, {"file_path": file_path})
        await self._job_scheduler.join()

        return self._job_manager.get_outputs(app_job)




from .read import ReadActivity
from .write import WriteActivity
from .generate import GenerateActivity
from .summarize import SummarizeActivity
from .extract import ExtractActivity
from .activity import ActivityError
from .store import StoreActivity
from .retrieve import RetrieveActivity
from .function import FunctionActivity
from .function import ReturnActivity

__all__ = [
    "ReadActivity",
    "WriteActivity",
    "GenerateActivity",
    "SummarizeActivity",
    "ExtractActivity",
    "ActivityError",
    "StoreActivity",
    "RetrieveActivity",
    "FunctionActivity",
    "ReturnActivity"
]



import logging
from typing import Dict, Any
from urllib.parse import urlencode

import httpx
from tenacity import retry, stop_after_attempt, wait_random_exponential


class AsyncHttpClient:
    TIMEOUT = 120

    def __init__(self):
        logging.getLogger('httpcore').setLevel(logging.ERROR)
        logging.getLogger('httpx').setLevel(logging.ERROR)

    @retry(stop=stop_after_attempt(5), wait=wait_random_exponential(min=1, max=60))
    async def post(self, url: str, headers: Dict[str, Any], data: Any, json=True) -> Any:
        async with httpx.AsyncClient() as ac:
            response = await ac.post(url, headers=headers, data=data, timeout=self.TIMEOUT)
        return response.json() if json else response.text

    async def get(self, url: str, query: Dict[str, Any] = None, headers: [str, Any] = None, json=True) -> Any:
        get_url = f"{url}?{urlencode(query)}" if query else url
        async with httpx.AsyncClient() as ac:
            response = await ac.get(get_url, headers=headers or {}, timeout=self.TIMEOUT)
        return response.json() if json else response.text

    @staticmethod
    def urljoin(*args):
        stripped = map(lambda x: str(x).strip('/'), args)
        return "/".join(stripped)




from typing import List
from pydantic import BaseModel
from typing import Dict, Any, Optional
from enum import Enum


class ActivityType(Enum):
    ANY = "any"
    READ = "read"
    WRITE = "write"
    STORE = "store"
    RETRIEVE = "retrieve"
    SUMMARIZE = "summarize"
    EXTRACT = "extract"
    GENERATE = "generate"
    FUNCTION = "function"
    CALL = "call"
    RETURN = "return"


class ToolType(Enum):
    ANY = "any"
    WEB = "web"
    REST = "rest"


class ModelProvider(Enum):
    ANY = "any"
    OPENAI = "openai"
    AZURE = "azure"
    ANTHROPIC = "anthropic"
    LLAVA = "llava"


class MemoryType(Enum):
    ANY = "any"
    CHROMADB = "chromadb"


class Model(BaseModel):
    model: str
    provider: ModelProvider
    parameters: Dict[str, Any] = {}


class ToolDef(BaseModel):
    type: ToolType
    models: Optional[List[str]] = None
    parameters: Dict[str, Any] = {}


class MemoryDef(BaseModel):
    type: MemoryType
    models: Optional[List[str]] = None
    parameters: Dict[str, Any] = {}


class ActivityInput(BaseModel):
    activity: str
    map: Optional[str] = None


class Activity(BaseModel):
    type: ActivityType
    tools: Optional[List[str]] = None
    inputs: Optional[List[ActivityInput]] = None
    models: Optional[List[str]] = None
    memory: Optional[List[str]] = None
    parameters: Dict[str, Any] = {}


class AppInfo(BaseModel):
    id: str
    title: str
    version: str
    profile: Optional[str] = None


class App(BaseModel):
    aq: str
    info: AppInfo
    models: Optional[Dict[str, Model]] = None
    memory: Optional[Dict[str, MemoryDef]] = None
    tools: Optional[Dict[str, ToolDef]] = None
    activities: Dict[str, Activity]




import os
import logging

from typing import List, Dict, Any

from ..repository import MemoryRepository
from ...types import MemoryDef

import chromadb
from chromadb.config import Settings
from ..splitter import split_into_chunks


class ChromaDbRepository(MemoryRepository):
    def __init__(self, config: Dict[str, Any]):
        self._config = config
        self._logger = logging.getLogger(self.__class__.__name__)
        logging.getLogger('chromadb').setLevel(logging.ERROR)

        path = self._config.get("path", "./data")
        if not os.path.exists(path):
            os.makedirs(path)

        self._chroma_client = chromadb.PersistentClient(path=path, settings=Settings(anonymized_telemetry=False))

    def store(self, memory_def: MemoryDef, collection: str, item_id: str, text: str, chunk_size: int) -> int:
        documents = []
        metas = []
        ids = []
        chunks = split_into_chunks(text, chunk_size)
        for i, chunk in enumerate(chunks):
            ids.append(f"{item_id} {i}")
            documents.append(chunk)
            metas.append({"item_id": item_id})

        collection = self._chroma_client.get_or_create_collection(collection)
        collection.add(ids=ids, documents=documents, metadatas=metas)
        self._logger.debug(f"Stored {len(chunks)} chunks")

        return len(ids)

    def retrieve(self, memory_def: MemoryDef, collection: str, query: str, n_results=3) -> List[str]:
        collection = self._chroma_client.get_or_create_collection(collection)
        results = collection.query(
            query_texts=[query],
            n_results=n_results
        )

        self._logger.debug(f"Retrieved {len(results['distances'][0])} chunks")
        for distance in results["distances"][0]:
            self._logger.debug(f"distance: {distance}")

        return results["documents"][0]



import logging
from typing import Dict, Any, List

from bs4 import BeautifulSoup
from pydantic import BaseModel

from ..tool import BaseTool, ToolError
from ...http_client import AsyncHttpClient
from ...providers.types import Tool
from ...types import ToolDef


class SearchResult(BaseModel):
    title: str
    snippet: str
    link: str


class SearchResponse(BaseModel):
    items: List[SearchResult]


class WebTool(BaseTool):
    USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36"

    def __init__(self, config: Dict[str, Any], http_client: AsyncHttpClient):
        self._config = config
        self._http_client = http_client
        self._logger = logging.getLogger(self.__class__.__name__)

    async def get_metadata(self, tool_def: ToolDef) -> List[Tool]:
        return [
            Tool(**{
                "function": {
                    "name": "search",
                    "description": "Searches the Internet and returns a list of matching web pages.",
                    "parameters": {
                        "required": ["query"],
                        "properties": {
                            "query": {
                                "description": "A search expression."
                            }
                        }
                    }
                }
            }),
            Tool(**{
                "function": {
                    "name": "summarize",
                    "description": "Summarizes the contents of a web page.",
                    "parameters": {
                        "required": ["link"],
                        "properties": {
                            "link": {
                                "description": "A universal resource locator (URL) pointing to a web page."
                            }
                        }
                    }
                }
            })
        ]

    async def invoke(self, function_name: str, arguments: Dict[str, Any], tool_def: ToolDef) -> str:
        if function_name == "search":
            if "query" not in arguments:
                raise ToolError("The query argument is required for search")
            response = await self.search(arguments["query"])
            return response.model_dump_json()
        elif function_name == "summarize":
            if "link" not in arguments:
                raise ToolError("The link argument is required for summarize")
            return await self.summarize(arguments["link"])
        else:
            raise ToolError(f"Unknown function {function_name}")

    async def search(self, query: str) -> SearchResponse:
        self._logger.debug(f"query = {query}")
        try:
            response = await self._http_client.get(self._config["endpoint"], {
                "cx": self._config["cx"],
                "key": self._config["key"],
                "q": query
            }, {
                "Accept": "application/json",
            })
            return SearchResponse(**response)
        except Exception as e:
            self._logger.error(f"Searched failed with error: {e}")

    async def summarize(self, link: str) -> str:
        self._logger.debug(f"link = {link}")

        if not link.startswith(("http://", "https://")):
            link = 'https://' + link.strip("/")

        html = await self._http_client.get(link, {}, {
            "User-Agent": self.USER_AGENT
        }, json=False)

        soup = BeautifulSoup(html, features="html.parser")
        text = soup.get_text()

        lines = (line.strip() for line in text.splitlines())
        chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
        text = '\n'.join(chunk for chunk in chunks if chunk)

        return text



import json
import re
import secrets
import string
from typing import Dict, Any

from jinja2 import Template
from jsonpath_ng import parse

from ..types import ActivityJob


def jsonpath(value: str, path: str) -> str:
    expr = parse(path)
    for match in expr.find(json.loads(value)):
        return match.value
    return ""


class ActivityError(Exception):
    pass


class BaseActivity:
    MAX_ITERATIONS = 42

    async def perform(self, activity_job: ActivityJob, inputs: Dict[str, Any]) -> None:
        pass

    @staticmethod
    def merge_inputs(inputs: Dict[str, Any]) -> str:
        values = []
        for key, val in inputs.items():
            if isinstance(val, list):
                values.append("\n".join(val))
            else:
                values.append(val)
        return "\n\n".join(values)

    @staticmethod
    def merge_inputs_json(inputs: Dict[str, Any], indent=2) -> str:
        rval = {}
        for key, val in inputs.items():
            try:
                rval[key] = [json.loads(elem) for elem in val] if isinstance(val, list) else json.loads(val)
            except json.JSONDecodeError:
                rval[key] = val
        if len(rval.keys()) == 1:
            return json.dumps(next(iter(rval.values())), indent=indent)
        else:
            return json.dumps(rval, indent=indent)

    @staticmethod
    def generate_temp_filename(prefix, extension, length=8):
        random_string = ''.join(secrets.choice(string.ascii_letters + string.digits) for _ in range(length))
        return f"{prefix}_{random_string}.{extension}"

    @staticmethod
    def render(template: str, inputs: Dict[str, Any]) -> str:
        # Call a function to process path expressions
        str_template = template
        expr = r'{{([^.\[]+)([.\[])(.*)}}'
        for match in re.finditer(expr, template):
            activity_name = match.group(1)
            start_of_expression = match.group(2)
            path_expression = match.group(3)
            str_template = str_template.replace(match.group(0),
                                                '{{' +
                                                f'jsonpath({activity_name}, "${start_of_expression}{path_expression}")'
                                                + '}}',
                                                1)

        # Render the template
        jinja_template = Template(str_template)
        jinja_template.globals.update({
            "jsonpath": jsonpath
        })
        return jinja_template.render(inputs)





from typing import List
from ..types import MemoryDef


class MemoryRepository:
    def store(self, memory_def: MemoryDef, collection: str, item_id: str, text: str, chunk_size: int) -> int:
        return 0

    def retrieve(self, memory_def: MemoryDef, collection: str, query: str, n_results=3) -> List[str]:
        return []




import logging
import time
from typing import Dict, Any

from .activity import BaseActivity, ActivityError
from ..types import ModelProvider, ActivityJob, JobState
from ..providers import ProviderManager
from ..providers.types import ChatCompletionMessage, ChatCompletionRequest, Choice


class SummarizeActivity(BaseActivity):
    def __init__(self, provider_manager: ProviderManager):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._provider_manager = provider_manager

    async def perform(self, activity_job: ActivityJob, inputs: Dict[str, Any]) -> None:
        try:
            app = activity_job.app_job.app
            activity = app.activities[activity_job.activity_name]

            # Get the text for summarization
            text = self.merge_inputs(inputs)
            if not text:
                raise ActivityError(f"Text is required")

            # Get the model
            if len(activity.models) < 1:
                raise ActivityError(f"A model is required")
            model = app.models[activity.models[0]]

            # Get the model parameters
            sentences = int(activity.parameters.get("sentences", model.parameters.get("sentences", 10)))
            temperature = float(activity.parameters.get("temperature", model.parameters.get("temperature", 0.5)))

            summary = await self.summarize(app.info.profile, text, model.provider,
                                           model.model, sentences, temperature)

            activity_job.output = summary
            activity_job.state = JobState.SUCCESS
            activity_job.output_type = "text/markdown"

        except Exception as e:
            activity_job.state = JobState.ERROR
            activity_job.output = str(e)
            self._logger.error(f"Encountered an error {e}")

    async def summarize(self, context: str, text: str,
                        provider_type: ModelProvider, model: str,
                        sentences: int, temperature: float) -> str:
        messages = []
        if context:
            messages.append(ChatCompletionMessage(role="system", content=context))

        prompt = """
            Summarize a block of text provided inside triple back ticks.
            Your summary must be readable. 
            Your summary must include about %(sentences)d sentences.
            ```%(text)s```
        """
        prompt = prompt % {"text": text, "sentences": sentences}
        messages.append(ChatCompletionMessage(role="user", content=prompt))

        request = ChatCompletionRequest(
            model=model,
            messages=messages,
            temperature=temperature
        )
        provider = self._provider_manager.get_provider(provider_type)

        start_time = time.perf_counter()

        response = await provider.create_completion(request)
        choice: Choice = response.choices[0]
        message: ChatCompletionMessage = choice.message

        self._logger.debug(f"Finished with reason {choice.finish_reason} "
                           f"in {int(time.perf_counter()-start_time)} sec.")

        return message.content



import json
import logging
import time
from typing import Dict, Any, List

from .activity import BaseActivity, ActivityError
from ..providers import ProviderManager
from ..providers.types import ChatCompletionMessage, ChatCompletionRequest, Choice, Tool, ResponseFormat, ToolCall
from ..tools import ToolManager
from ..types import ActivityJob, JobState, Activity, App


class GenerateActivity(BaseActivity):
    TOOL_NAME_DELIMITER = "__"

    def __init__(self, provider_manager: ProviderManager, tool_manager: ToolManager):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._provider_manager = provider_manager
        self._tool_manager = tool_manager

    async def perform(self, activity_job: ActivityJob, inputs: Dict[str, Any]) -> None:
        try:
            app = activity_job.app_job.app
            activity = app.activities[activity_job.activity_name]

            if len(activity.models) < 1:
                raise ActivityError(f"A model is required")
            model = app.models[activity.models[0]]

            temperature = float(activity.parameters.get("temperature", model.parameters.get("temperature", 0.5)))
            max_tokens = int(activity.parameters.get("max_words", model.parameters.get("max_words", 500))*4/3)

            messages = []
            profile = app.info.profile
            if profile:
                messages.append(ChatCompletionMessage(role="system", content=profile))

            json_format = activity.parameters.get("format", None) == "json"
            if json_format:
                messages.append(ChatCompletionMessage(
                    role="system",
                    content="Provide your response as a JSON object."))
            else:
                messages.append(ChatCompletionMessage(
                    role="system",
                    content="Use the tab length of two spaces when formatting nested lists in markdown."))

            tools = await self.get_tools(app, activity)
            if tools:
                messages.append(ChatCompletionMessage(
                    role="system",
                    content="Think step-by-step. Perform as many iterations as necessary "
                            "to accomplish your goal using the tools provided."))

            prompt_template = activity.parameters["prompt"]
            prompt = self.render(prompt_template, inputs)
            messages.append(ChatCompletionMessage(role="user", content=prompt))

            parts = []
            start_time = time.perf_counter()
            provider = self._provider_manager.get_provider(model.provider)
            for x in range(self.MAX_ITERATIONS):
                request = ChatCompletionRequest(
                    model=model.model,
                    messages=messages,
                    temperature=temperature,
                    max_tokens=max_tokens,
                    tools=tools if tools else None,
                    tool_choice="auto" if tools else None,
                    response_format=ResponseFormat(type="json_object") if json_format else None
                )

                response = await provider.create_completion(request)

                choice: Choice = response.choices[0]
                message: ChatCompletionMessage = choice.message
                messages.append(message)

                if choice.finish_reason == "tool_calls":
                    for tool_call in message.tool_calls:
                        tool_result = await self.process_tool_call(tool_call, app)
                        messages.append(tool_result)
                else:
                    if message.content:
                        parts.append(message.content)
                    if choice.finish_reason:
                        self._logger.debug(f"Finished with reason {choice.finish_reason} "
                                           f"in {int(time.perf_counter()-start_time)} sec.")
                        break

            activity_job.state = JobState.SUCCESS
            activity_job.output = "\n\n".join(parts)
            activity_job.output_type = "text/markdown"

        except Exception as e:
            self._logger.error(e)
            activity_job.state = JobState.ERROR
            activity_job.output = str(e)

    async def get_tools(self, app: App, activity: Activity) -> List[Tool]:
        tools = []
        if activity.tools:
            for tool_name in activity.tools:
                tool_def = app.tools[tool_name]
                tool_obj = self._tool_manager.get_tool(tool_def.type)
                metadata = await tool_obj.get_metadata(tool_def)
                for tool in metadata:
                    func = tool.function
                    func.name = f"{tool_name}{self.TOOL_NAME_DELIMITER}{func.name}"
                    tools.append(tool)
        return tools

    async def process_tool_call(self, tool_call: ToolCall, app: App) -> ChatCompletionMessage:
        self._logger.debug(f"Calling {tool_call.function.name}")

        names = tool_call.function.name.split(self.TOOL_NAME_DELIMITER)
        if len(names) < 2:
            raise ActivityError(f"Invalid tool name {tool_call.function.name}")

        if names[0] not in app.tools:
            raise ActivityError(f"{names[0]} is not a valid tool name")

        tool_def = app.tools[names[0]]
        tool_obj = self._tool_manager.get_tool(tool_def.type)

        arguments = json.loads(tool_call.function.arguments)
        response = await tool_obj.invoke(names[1], arguments, tool_def)

        return ChatCompletionMessage(
            role="tool",
            tool_call_id=tool_call.id,
            name=tool_call.function.name,
            content=response
        )



import logging
from typing import Dict, Any

from pydantic import BaseModel

from ..provider import BaseProvider, ProviderError
from ..types import ChatCompletionRequest, ChatCompletionResponse, Choice, ChatCompletionMessage
from ...http_client import AsyncHttpClient


class AnthropicCompletionRequest(BaseModel):
    model: str
    prompt: str
    temperature: float
    max_tokens_to_sample: int


class AnthropicCompletionResponse(BaseModel):
    completion: str
    model: str
    stop_reason: str


class AnthropicProvider(BaseProvider):
    def __init__(self, config: Dict[str, Any], http_client: AsyncHttpClient):
        self._config = config
        self._http_client = http_client
        self._logger = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def _check_config(config: Dict[str, Any]) -> None:
        required_keys = ['endpoint', 'key']
        if not all(key in config for key in required_keys):
            raise ProviderError("The Anthropic provider is not configured. Add settings to config.yml.")

    async def create_completion(self, request: ChatCompletionRequest) -> ChatCompletionResponse:
        self._check_config(self._config)

        # Format the prompt
        system_prompts = []
        prompts = []
        for message in request.messages:
            if message.role == "system":
                system_prompts.append(message.content)
            elif message.role == "user":
                prompts.append(f"Human: {message.content}")
            elif message.role == "assistant":
                prompts.append(f"Assistant: {message.content}")
            else:
                self._logger.error(f"Unknown role {message.role}")

        system_prompts.append("")
        prompts.append("Assistant:")
        prompt = "\n".join(system_prompts) + "\n\n" + "\n\n".join(prompts)

        # Map and send the request
        a_request = AnthropicCompletionRequest(model=request.model,
                                               prompt=prompt,
                                               temperature=request.temperature,
                                               max_tokens_to_sample=request.max_tokens)
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "x-api-key": self._config['key'],
            "anthropic-version": "2023-06-01"
        }
        url = self._http_client.urljoin(self._config["endpoint"], "/complete")
        data = await self._http_client.post(url, headers, a_request.model_dump_json(exclude_none=True))

        # Map and return the response
        a_response = AnthropicCompletionResponse(**data)
        finish_reason = "stop" if a_response.stop_reason == "stop_sequence" else a_response.stop_reason
        message = ChatCompletionMessage(role="assistant", content=a_response.completion)
        choice = Choice(index=0, message=message, finish_reason=finish_reason)
        response = ChatCompletionResponse(id="", object="object", created=0, choices=[choice])

        return response



import uuid
from enum import Enum
from typing import Dict, Optional

from .app import App


class JobState(Enum):
    ANY = 0,
    CREATED = 1,
    RUNNING = 2,
    SUCCESS = 3
    ERROR = 4


class AppJob:
    id: str
    caller: Optional['ActivityJob']
    app: App
    context: Dict
    state: JobState

    def __init__(self, app: App):
        self.id = str(uuid.uuid4())
        self.app = app
        self.context = {}
        self.caller = None
        self.state = JobState.CREATED

    @property
    def finished(self) -> bool:
        return self.state == JobState.SUCCESS or self.state == JobState.ERROR


class ActivityJob:
    id: str
    activity_name: str
    app_job: AppJob
    state: JobState
    output: str
    output_type: str = "text/plain"

    def __init__(self, activity_name: str, app_job: AppJob):
        self.id = str(uuid.uuid4())
        self.activity_name = activity_name
        self.app_job = app_job
        self.state = JobState.CREATED
        self.output = ""

    @property
    def finished(self) -> bool:
        return self.state == JobState.SUCCESS or self.state == JobState.ERROR







from .app import (
    App,
    Activity,
    ActivityType,
    ActivityInput,
    ModelProvider,
    Model,
    ToolDef,
    ToolType,
    MemoryType,
    MemoryDef)

from .job import ActivityJob, AppJob, JobState


__all__ = [
    "ActivityType",
    "ModelProvider",
    "Model",
    "ToolDef",
    "ToolType",
    "ActivityInput",
    "Activity",
    "App",
    "ActivityJob",
    "AppJob",
    "JobState",
    "MemoryType",
    "MemoryDef"
]



import logging
from typing import Dict, Any

from .activity import BaseActivity
from ..types import ActivityJob, JobState


class FunctionActivity(BaseActivity):
    def __init__(self):
        self._logger = logging.getLogger(self.__class__.__name__)

    async def perform(self, activity_job: ActivityJob, inputs: Dict[str, Any]) -> None:
        activity_job.state = JobState.SUCCESS
        activity_job.output = self.merge_inputs(inputs)
        activity_job.output_type = "text/plain"


class ReturnActivity(BaseActivity):
    def __init__(self):
        self._logger = logging.getLogger(self.__class__.__name__)

    async def perform(self, activity_job: ActivityJob, inputs: Dict[str, Any]) -> None:
        try:
            activity_job.state = JobState.SUCCESS
            activity_job.output = self.merge_inputs_json(inputs)
            activity_job.output_type = "application/json"
        except Exception as e:
            activity_job.state = JobState.ERROR
            activity_job.output = str(e)
            self._logger.error(f"Encountered an error {e}")




import json
import logging
from typing import Dict, List, Any

from jsonpath_ng import parse

from ..types import App, Activity, AppJob, ActivityJob


class AppJobError(Exception):
    pass


class JobManager:
    app_jobs: Dict[str, AppJob]
    activity_jobs: Dict[str, Dict[str, List[ActivityJob]]]

    def __init__(self):
        self.app_jobs = {}
        self.activity_jobs = {}
        self._logger = logging.getLogger(self.__class__.__name__)

    def create_app_job(self, app: App) -> AppJob:
        app_job = AppJob(app)
        self.app_jobs[app_job.id] = app_job
        self.activity_jobs[app_job.id] = {}
        return app_job

    def create_activity_job(self, app_job: AppJob, activity_name: str) -> ActivityJob:
        app = app_job.app
        if activity_name not in app.activities:
            raise AppJobError(f"Activity {activity_name} not found")
        activity_job = ActivityJob(activity_name, app_job)
        self.activity_jobs[app_job.id].setdefault(activity_name, []).append(activity_job)
        return activity_job

    def get_next_activities(self, activity_job: ActivityJob) -> List[str]:
        app = activity_job.app_job.app
        rv = []
        for activity_name in app.activities:
            activity = app.activities[activity_name]
            for activity_input in activity.inputs or []:
                if (activity_input.activity == activity_job.activity_name and
                        not self.is_waiting_for_jobs(activity_job.app_job, activity)):
                    rv.append(activity_name)
        return rv

    def is_waiting_for_jobs(self, app_job: AppJob, activity: Activity):
        for activity_input in activity.inputs or []:
            jobs = self.activity_jobs[app_job.id].get(activity_input.activity, None)
            if not jobs or any(not job.finished for job in jobs):
                return True
        return False

    def get_inputs_for_activity(self, app_job: AppJob, activity: Activity) -> List[Dict[str, Any]]:
        inputs_for_activity = {}
        if activity.inputs:
            for activity_input in activity.inputs:
                job_outputs = [job.output for job in self.activity_jobs[app_job.id].get(activity_input.activity, [])]
                inputs_for_activity[activity_input.activity] = job_outputs if len(job_outputs) > 1 else job_outputs[0]

            for activity_input in activity.inputs:
                if activity_input.map:
                    try:
                        expr = parse(activity_input.map)
                        val = inputs_for_activity[activity_input.activity]
                        inputs = []
                        for match in expr.find(json.loads(val)):
                            if isinstance(match.value, list):
                                for input_value in match.value:
                                    inputs.append({**inputs_for_activity, activity_input.activity: input_value})
                        return inputs
                    except Exception as e:
                        self._logger.error(f"Failed to parse a map expression {e}")
                        return []

        return [inputs_for_activity]

    def get_outputs(self, app_job: AppJob) -> Dict[str, Any]:
        app = app_job.app

        # Terminal activities have inputs and do not have any other activities that take their outputs
        terminal_activities = [
            activity_name
            for activity_name, activity in app.activities.items()
            if activity.inputs and not any(
                activity_input.activity == activity_name
                for some_other_activity in app.activities.values()
                for activity_input in some_other_activity.inputs or []
            )
        ]

        # Collect outputs from terminal activities
        outputs = {
            activity_name: [job.output for job in self.activity_jobs[app_job.id].get(activity_name, [])]
            for activity_name in terminal_activities
        }

        # Remove empty values and return
        return {key: value for key, value in outputs.items() if value}



from typing import List, Dict, Any

from ..providers.types import Tool
from ..types import ToolDef


class ToolError(Exception):
    pass


class BaseTool:
    async def get_metadata(self, tool_def: ToolDef) -> List[Tool]:
        pass

    async def invoke(self, function_name: str, arguments: Dict[str, Any], tool_def: ToolDef) -> str:
        pass



from .chat import (
    ChatCompletionMessage,
    ChatCompletionRequest,
    ChatCompletionResponse,
    Choice,
    Parameters,
    Property,
    Function,
    Tool,
    ResponseFormat,
    ToolCall,
    Error
)

__all__ = [
    "ChatCompletionMessage",
    "ChatCompletionRequest",
    "ChatCompletionResponse",
    "Choice",
    "Parameters",
    "Property",
    "Function",
    "Tool",
    "ResponseFormat",
    "ToolCall",
    "Error"
]



import asyncio
import logging
import time
from typing import Dict, Any

from .manager import JobManager, AppJobError
from ..activities import (
    ReadActivity,
    WriteActivity,
    SummarizeActivity,
    GenerateActivity,
    ExtractActivity,
    StoreActivity,
    RetrieveActivity,
    FunctionActivity,
    ReturnActivity
)
from ..types import ActivityType, ActivityJob, JobState


class WorkItem:
    job: ActivityJob
    inputs: Dict[str, Any]

    def __init__(self, job: ActivityJob, inputs: Dict[str, Any]) -> None:
        self.job = job
        self.inputs = inputs


class JobScheduler:
    def __init__(self, config: Dict[str, Any], job_manager: JobManager,
                 read_activity: ReadActivity, write_activity: WriteActivity,
                 summarize_activity: SummarizeActivity, generate_activity: GenerateActivity,
                 extract_activity: ExtractActivity, store_activity: StoreActivity, retrieve_activity: RetrieveActivity,
                 function_activity: FunctionActivity, return_activity: ReturnActivity):
        self._config = config

        self._logger = logging.getLogger(self.__class__.__name__)
        logging.getLogger('asyncio').setLevel(logging.ERROR)

        self._queue = asyncio.Queue()
        self._workers = []
        self._job_manager = job_manager

        self._activity_handlers = {
            ActivityType.READ: read_activity,
            ActivityType.WRITE: write_activity,
            ActivityType.SUMMARIZE: summarize_activity,
            ActivityType.GENERATE: generate_activity,
            ActivityType.EXTRACT: extract_activity,
            ActivityType.STORE: store_activity,
            ActivityType.RETRIEVE: retrieve_activity,
            ActivityType.FUNCTION: function_activity,
            ActivityType.RETURN: return_activity
        }

    async def start_workers(self):
        num_workers = self._config.get("workers", 3)
        self._workers = [asyncio.create_task(self.consume(n)) for n in range(num_workers)]

    async def schedule(self, activity_job: ActivityJob, inputs: Dict[str, Any]) -> None:
        item = WorkItem(activity_job, inputs)
        await self._queue.put(item)

    async def consume(self, name: int) -> None:
        while True:
            item = await self._queue.get()
            self._logger.debug(f"Worker {name} performing {item.job.activity_name}")
            start_time = time.perf_counter()
            await self.perform(item)
            self._logger.debug(f"Finished {item.job.activity_name} in {int(time.perf_counter()-start_time)} sec.")
            self._queue.task_done()

    async def perform(self, item: WorkItem) -> None:
        item.job.state = JobState.RUNNING
        app = item.job.app_job.app
        app_job = item.job.app_job

        activity = app.activities[item.job.activity_name]
        activity_type = activity.type

        if activity_type == ActivityType.CALL:
            # Push a new app job on the call stack
            function_name = activity.parameters["function"]
            if function_name and function_name in app.activities:
                app_job = self._job_manager.create_app_job(app)
                app_job.caller = item.job
                function_job = self._job_manager.create_activity_job(app_job, function_name)
                await self.schedule(function_job, item.inputs)
            else:
                raise AppJobError(f"The function name in a call activity is incorrect or missing")
        else:
            handler = self._activity_handlers.get(activity_type)
            if handler:
                await handler.perform(item.job, item.inputs)
                if item.job.state == JobState.SUCCESS:
                    # Pop the app job stack if it's a return activity
                    if activity_type == ActivityType.RETURN and app_job.caller:
                        # Update the state of the caller activity
                        activity_job = app_job.caller
                        activity_job.state = JobState.SUCCESS
                        activity_job.output = item.job.output
                        activity_job.output_type = item.job.output_type

                        # Update the state of the callee app job
                        app_job.state = JobState.SUCCESS
                        app_job = activity_job.app_job
                    else:
                        activity_job = item.job

                    # Schedule next jobs
                    next_activities = self._job_manager.get_next_activities(activity_job)
                    for next_activity in next_activities:
                        all_inputs = self._job_manager.get_inputs_for_activity(app_job, app.activities[next_activity])
                        for job_inputs in all_inputs:
                            next_job = self._job_manager.create_activity_job(app_job, next_activity)
                            await self.schedule(next_job, job_inputs)
                else:
                    self._logger.error(f"{item.job.activity_name} failed with error {item.job.output}")
            else:
                raise AppJobError(f"Unknown activity type {activity_type}")

    async def join(self) -> None:
        await self._queue.join()
        for w in self._workers:
            w.cancel()
        self._logger.debug("Finished work items")



import logging
import uuid
from typing import Dict, Any

from .activity import BaseActivity, ActivityError
from ..types import ActivityJob, JobState
from ..memory import MemoryManager


class StoreActivity(BaseActivity):
    def __init__(self, memory_manager: MemoryManager):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._memory_manager = memory_manager

    async def perform(self, activity_job: ActivityJob, inputs: Dict[str, Any]) -> None:
        try:
            app = activity_job.app_job.app
            if not app.memory:
                raise ActivityError("A memory repository is required for this app")

            activity = app.activities[activity_job.activity_name]
            if not activity.memory:
                raise ActivityError("A memory repository is required for this activity")

            memory_def = app.memory[activity.memory[0]]
            memory_repository = self._memory_manager.get_repository(memory_def.type)

            file_id = activity_job.app_job.context.get("file_path", str(uuid.uuid4()))
            chunk_size = activity.parameters.get("chunk_size", memory_def.parameters.get("chunk_size", 2000))

            text = self.merge_inputs(inputs)
            chunks = memory_repository.store(memory_def, app.info.id, file_id, text, chunk_size)

            activity_job.state = JobState.SUCCESS
            activity_job.output = str(chunks)
            activity_job.output_type = "text/plain"

        except Exception as e:
            activity_job.state = JobState.ERROR
            activity_job.output = str(e)
            self._logger.error(f"Encountered an error {e}")



from .reader import Reader, ReaderError
import pypdf


class PdfReader(Reader):
    async def read(self, file_path: str) -> str:
        try:
            reader = pypdf.PdfReader(file_path)
            text = [page.extract_text() for page in reader.pages]
            return "\n\n".join(text)
        except Exception as e:
            raise ReaderError(f"Failed to read text from PDF: {e}")



from typing import Optional, List, Dict, Literal

from pydantic import BaseModel


class FunctionCall(BaseModel):
    name: str
    arguments: str


class ToolCall(BaseModel):
    id: str
    type: str
    function: FunctionCall


class ChatCompletionMessage(BaseModel):
    role: str
    content: Optional[str] = ""
    name: Optional[str] = None
    tool_call_id: Optional[str] = None
    tool_calls: Optional[List[ToolCall]] = None


class Property(BaseModel):
    type: Literal["string"] = "string"
    description: str
    enum: Optional[List[str]] = None


class Parameters(BaseModel):
    type: Literal["object"] = "object"
    properties: Dict[str, Property]
    required: List[str]


class Function(BaseModel):
    name: str
    description: str
    parameters: Parameters


class Tool(BaseModel):
    type: Literal["function"] = "function"
    function: Function


class ResponseFormat(BaseModel):
    type: Literal["json_object"] = "json_object"


class ChatCompletionRequest(BaseModel):
    model: str
    messages: List[ChatCompletionMessage]
    tools: Optional[List[Tool]] = None
    response_format: Optional[ResponseFormat] = None
    tool_choice: Optional[str] = None
    temperature: float = 0.5
    presence_penalty: float = 0.0
    frequency_penalty: float = 0.0
    max_tokens: int = 1000


class Choice(BaseModel):
    index: int
    message: ChatCompletionMessage
    finish_reason: Optional[str] = None


class ChatCompletionResponse(BaseModel):
    id: str
    object: str
    created: int
    choices: List[Choice]


class Error(BaseModel):
    code: str
    message: str






from .types import ChatCompletionRequest, ChatCompletionResponse


class ProviderError(Exception):
    def __init__(self, code, message):
        self.code = code
        super().__init__(message)


class BaseProvider:
    async def create_completion(self, request: ChatCompletionRequest) -> ChatCompletionResponse:
        pass


import logging

from .openai import OpenAIProvider
from .azure import AzureProvider
from .anthropic import AnthropicProvider
from .llava import LlavaProvider
from ..types import ModelProvider
from .provider import BaseProvider


class ProviderManager:
    def __init__(self, openai_provider: OpenAIProvider,
                 azure_provider: AzureProvider,
                 anthropic_provider: AnthropicProvider,
                 llava_provider: LlavaProvider):
        self._providers = {
            ModelProvider.OPENAI: openai_provider,
            ModelProvider.AZURE: azure_provider,
            ModelProvider.ANTHROPIC: anthropic_provider,
            ModelProvider.LLAVA: llava_provider
        }
        self._logger = logging.getLogger(self.__class__.__name__)

    def get_provider(self, provider_type: ModelProvider) -> BaseProvider:
        return self._providers[provider_type]




import logging
from typing import Dict, Any

from ..provider import BaseProvider, ProviderError
from ...http_client import AsyncHttpClient
from ..types import ChatCompletionRequest, ChatCompletionResponse, Error


class OpenAIProvider(BaseProvider):
    def __init__(self, config: Dict[str, Any], http_client: AsyncHttpClient):
        self._config = config
        self._http_client = http_client
        self._logger = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def _check_config(config: Dict[str, Any]) -> None:
        required_keys = ['endpoint', 'key']
        if not all(key in config for key in required_keys):
            raise ProviderError("The OpenAI provider is not configured. Add settings to config.yml.")

    async def create_completion(self, request: ChatCompletionRequest) -> ChatCompletionResponse:
        self._check_config(self._config)

        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._config['key']}"
        }
        url = self._http_client.urljoin(self._config["endpoint"], "/chat/completions")
        response = await self._http_client.post(url, headers, request.model_dump_json(exclude_none=True))

        if "error" in response:
            error = Error(**response['error'])
            raise ProviderError(error.code, error.message)
        else:
            return ChatCompletionResponse(**response)



import logging
import os
from typing import Dict, Any

import aiofiles
import markdown

from .activity import BaseActivity
from ..types import ActivityJob, JobState


class WriteActivity(BaseActivity):
    HTML_TEMPLATE = """<html>
<head>
    <meta http-equiv="content-type" content="text/html; charset=UTF-8">
    <style>
        .content {
            width: 80%%;
            margin: auto;
            line-height: 1.4rem;
            font-family: Helvetica, Arial, sans-serif;
            font-size: 0.9rem;
        }
    </style>
</head>
<body>
    <div class="content">
        %s
    </div>
</body>
</html>"""

    def __init__(self):
        self._logger = logging.getLogger(self.__class__.__name__)

    async def perform(self, activity_job: ActivityJob, inputs: Dict[str, Any]) -> None:
        try:
            app = activity_job.app_job.app
            activity = app.activities[activity_job.activity_name]

            # Collate the inputs
            output_format = activity.parameters.get("format", "md")
            template = activity.parameters.get("template", None)
            if template:
                text = self.render(template, inputs)
            elif output_format == "json":
                text = self.merge_inputs_json(inputs, indent=2)
            else:
                text = self.merge_inputs(inputs)

            # Compute the file prefix based on the original file name
            original_file_path = activity_job.app_job.context.get("file_path", None)
            if original_file_path:
                base_name = os.path.basename(original_file_path)
                file_prefix, _ = os.path.splitext(base_name)
            else:
                file_prefix = "out"

            # Apply formatting
            if output_format == "html":
                text = self.HTML_TEMPLATE % markdown.markdown(text, tab_length=2)

            # Write content to file
            file_name = self.generate_temp_filename(file_prefix, output_format)

            # Create the out directory
            path = "./out"
            if not os.path.exists(path):
                os.makedirs(path)

            file_path = os.path.join(path, file_name)
            async with aiofiles.open(file_path, mode='w', encoding='utf-8') as file:
                await file.write(text)

            activity_job.state = JobState.SUCCESS
            activity_job.output = file_path

        except Exception as e:
            self._logger.error(e)
            activity_job.state = JobState.ERROR
            activity_job.output = str(e)







import logging

from .repository import MemoryRepository
from .chromadb import ChromaDbRepository
from ..types import MemoryType


class MemoryManager:
    def __init__(self, chromadb_repository: ChromaDbRepository):
        self._repositories = {
            MemoryType.CHROMADB: chromadb_repository
        }
        self._logger = logging.getLogger(self.__class__.__name__)

    def get_repository(self, memory_type: MemoryType) -> MemoryRepository:
        return self._repositories[memory_type]



from .web import WebTool
from .rest import RestTool
from .tool import BaseTool
from ..types import ToolType


class ToolManager:
    def __init__(self, web_tool: WebTool, rest_tool: RestTool):
        self._tools = {
            ToolType.WEB: web_tool,
            ToolType.REST: rest_tool
        }

    def get_tool(self, tool_type: ToolType) -> BaseTool:
        return self._tools.get(tool_type, None)



import re

alphabets= "([A-Za-z])"
prefixes = "(Mr|St|Mrs|Ms|Dr)[.]"
suffixes = "(Inc|Ltd|Jr|Sr|Co)"
starters = "(Mr|Mrs|Ms|Dr|Prof|Capt|Cpt|Lt|He\s|She\s|It\s|They\s|Their\s|Our\s|We\s|But\s|However\s|That\s|This\s|Wherever)"
acronyms = "([A-Z][.][A-Z][.](?:[A-Z][.])?)"
websites = "[.](com|net|org|io|gov|edu|me)"
digits = "([0-9])"
multiple_dots = r'\.{2,}'


def split_into_sentences(text: str) -> list[str]:
    text = " " + text + "  "
    text = text.replace("\n"," ")
    text = re.sub(prefixes,"\\1<prd>",text)
    text = re.sub(websites,"<prd>\\1",text)
    text = re.sub(digits + "[.]" + digits,"\\1<prd>\\2",text)
    text = re.sub(multiple_dots, lambda match: "<prd>" * len(match.group(0)) + "<stop>", text)
    if "Ph.D" in text: text = text.replace("Ph.D.","Ph<prd>D<prd>")
    text = re.sub("\s" + alphabets + "[.] "," \\1<prd> ",text)
    text = re.sub(acronyms+" "+starters,"\\1<stop> \\2",text)
    text = re.sub(alphabets + "[.]" + alphabets + "[.]" + alphabets + "[.]","\\1<prd>\\2<prd>\\3<prd>",text)
    text = re.sub(alphabets + "[.]" + alphabets + "[.]","\\1<prd>\\2<prd>",text)
    text = re.sub(" "+suffixes+"[.] "+starters," \\1<stop> \\2",text)
    text = re.sub(" "+suffixes+"[.]"," \\1<prd>",text)
    text = re.sub(" " + alphabets + "[.]"," \\1<prd>",text)
    if "”" in text: text = text.replace(".”","”.")
    if "\"" in text: text = text.replace(".\"","\".")
    if "!" in text: text = text.replace("!\"","\"!")
    if "?" in text: text = text.replace("?\"","\"?")
    text = text.replace(".",".<stop>")
    text = text.replace("?","?<stop>")
    text = text.replace("!","!<stop>")
    text = text.replace("<prd>",".")
    sentences = text.split("<stop>")
    sentences = [s.strip() for s in sentences]
    if sentences and not sentences[-1]: sentences = sentences[:-1]
    return sentences


def split_into_chunks(text: str, size: int) -> list[str]:
    chunks = []
    chunk = []
    chunk_len = 0
    sentences = split_into_sentences(text)
    for sentence in sentences:
        if chunk_len + len(sentence) >= size:
            chunks.append(" ".join(chunk))
            chunk_len = len(sentence) + 1
            chunk = [sentence]
        else:
            chunk.append(sentence)
            chunk_len += len(sentence) + 1
    chunks.append(" ".join(chunk))
    return chunks



import asyncio
import sys

from aq.containers import get_broker


if __name__ == "__main__":
    if len(sys.argv) <= 1:
        print("Usage: python " + sys.argv[0] + " <app file> <activity name> <file path>")
    else:
        broker = get_broker()
        rv = asyncio.run(broker.run(*sys.argv[1:]))
        print(rv)



import logging
from typing import Dict, Any

from .activity import BaseActivity, ActivityError
from ..types import ActivityJob, JobState
from ..memory import MemoryManager


class RetrieveActivity(BaseActivity):
    def __init__(self, memory_manager: MemoryManager):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._memory_manager = memory_manager

    async def perform(self, activity_job: ActivityJob, inputs: Dict[str, Any]) -> None:
        try:
            app = activity_job.app_job.app
            if not app.memory:
                raise ActivityError("A memory repository is required for this app")

            activity = app.activities[activity_job.activity_name]
            if not activity.memory:
                raise ActivityError("A memory repository is required for this activity")

            memory_def = app.memory[activity.memory[0]]
            memory_repository = self._memory_manager.get_repository(memory_def.type)

            n_results = activity.parameters.get("n_results", 3)

            chunks = memory_repository.retrieve(memory_def, app.info.id, self.merge_inputs(inputs), n_results)

            activity_job.state = JobState.SUCCESS
            activity_job.output = "\n\n".join(chunks)
            activity_job.output_type = "text/plain"

        except Exception as e:
            activity_job.state = JobState.ERROR
            activity_job.output = str(e)
            self._logger.error(f"Encountered an error {e}")
